/*  Autopackage GTK+ frontend
 *  Copyright (C) 2003  Hongli Lai <h.lai@chello.nl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include "processor.h"
#include "protocol.h"
#include "utils.h"
#include "main.h"
#include "receiver.h"
#include "interactions.h"
#include "logview.h"
#include "eggstatusicon.h"
#include "gul-ellipsizing-label.h"

#define W(x) glade_xml_get_widget (xml, x)

#define PROTOCOL_VERSION "4"

#define true TRUE

static gboolean helloInitialized = FALSE;
static gboolean failed = FALSE;
static guint sessionDepth = 1;
static gboolean installStarted = FALSE;
static gboolean progressStarted = FALSE;
static guint maxPackages = 0;
static guint packagesInstalled = 0;
static GList *interactionList = NULL;
static gboolean summaryShown = FALSE;

#define COMPLETED_PROBLEM 0
#define COMPLETED_INSTALL 1
#define COMPLETED_UNINSTALL 2
#define COMPLETED_VERIFY 3
#define COMPLETED_UNKNOWN 4
static gboolean presented = FALSE;
static void present_completed (int what);


static gboolean
process_hello (gchar *data)
{
	gchar *first, *tmp;

	/* Initial handshake for every package */

	tmp = strchr (data, '\n');
	first = g_strndup (data, tmp - data);

	if (my_strcmp (first + 6, PROTOCOL_VERSION))
	{
		/* Don't increase the session depth if this is the first HELLO. */
		if (helloInitialized)
			sessionDepth++;

		protocol_send ("LETS GO");
		g_free (first);
		helloInitialized = TRUE;

		return TRUE;
	} else
	{
		protocol_send ("UHOH VERSION_MISMATCH");
		show_error (W("main"),
			_("Protocol version mismatched (expected %s, but got %s).\n"
			"Please ensure your installation of autopackage is correct."),
			PROTOCOL_VERSION, first + 6);
		g_free (first);
		return FALSE;
	}
}


static gboolean
process_terminate ()
{
	sessionDepth--;
	if (sessionDepth > 0)
		return TRUE;

	/* The backend terminated */

	pulse_stop ("globalProgress");

	/* We exited normally so no need to remove the FIFO */
	pokeFifo = FALSE;
	receiver_stop ();

	if (failed)
		present_completed (COMPLETED_PROBLEM);
	else if (!presented)
		present_completed (COMPLETED_UNKNOWN);

	finished = TRUE;
	return TRUE;
}


static gboolean
process_switch_group (gchar *data)
{
	/* Ignore this */
	return TRUE;
}


static gboolean
process_identify (gchar *data)
{
	GList *lines;
	gchar *displayName;
	guint id;
	gchar *context;
	gchar *msg;

	/* We are current processing a package */

	lines = create_strlist (data);
	displayName = (gchar *) g_list_nth_data (lines, 1);
	id = atoi ((gchar *) g_list_nth_data (lines, 2));
	context = (gchar *) g_list_nth_data (lines, 3);

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);

	gtk_label_set_text (GTK_LABEL (W("name")), displayName);
	if (my_strcmp (context, "prep"))
	{
		GList *children;

		/* Preparing a package (checking dependancies) */
		gtk_label_set_markup (GTK_LABEL (W("title")),
			_("<span weight=\"bold\" size=\"x-large\">Preparing package:</span>"));
		gtk_widget_show (W("prepareImage"));
		gtk_widget_hide (W("installImage"));
		gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);

		children = gtk_container_get_children (GTK_CONTAINER (
				GTK_LOG_VIEW (W("logview"))->vbox
			));
		if (g_list_first (children))
			gtk_log_view_add_separator (GTK_LOG_VIEW (W("logview")));
		g_list_free (children);

		set_title (_("Preparing package: %s"), displayName);

		if (!failed) gtk_log_view_clear (GTK_LOG_VIEW (W("logview")));

	} else if (my_strcmp (context, "install"))
	{
		/* Installing a package */

		gtk_label_set_markup (GTK_LABEL (W("title")),
			_("<span weight=\"bold\" size=\"x-large\">Currently installing:</span>"));
		gtk_widget_hide (W("prepareImage"));
		gtk_widget_show (W("installImage"));

		packagesInstalled = id;
		gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
		gtk_log_view_clear (GTK_LOG_VIEW (W("logview")));

		if (maxPackages > 1)
		{
			set_title (_("Installing: %s (package %d of %d)"),
				displayName, id, maxPackages,
				((gdouble) id / (gdouble) (maxPackages + 1)) * 100.0);

			msg = g_strdup_printf ("Package %d of %d", id, maxPackages);
			gtk_progress_bar_set_text (GTK_PROGRESS_BAR (W("globalProgress")), msg);
			g_free (msg);
			gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (W("globalProgress")),
				(gdouble) id / (gdouble) (maxPackages + 1));
		} else
			set_title (_("Installing: %s"), displayName);
	} else if (my_strcmp (context, "verify"))
	{
		/* verifying a package */
		char *msg = g_strdup_printf("<span weight=\"bold\" size=\"x-large\">%s:</span>", _("Currently verifying"));
		gtk_label_set_markup (GTK_LABEL (W("title")), msg);
		g_free(msg);
		gtk_widget_hide (W("prepareImage"));
		gtk_widget_show (W("installImage"));

		gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
		gtk_log_view_clear (GTK_LOG_VIEW (W("logview")));

		if (maxPackages > 1)
		{
			set_title (_("Verify: %s (package %d of %d)"),
				displayName, id, maxPackages,
				((gdouble) id / (gdouble) (maxPackages + 1)) * 100.0);

			msg = g_strdup_printf ("Package %d of %d", id, maxPackages);
			gtk_progress_bar_set_text (GTK_PROGRESS_BAR (W("globalProgress")), msg);
			g_free (msg);
			gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (W("globalProgress")),
				(gdouble) id / (gdouble) (maxPackages + 1));
		} else
			set_title (_("Verifying %s"), displayName);
		
	}

	free_strlist (lines);
	return TRUE;
}


static gboolean
process_series (gchar *data)
{
	gchar *max;

	/* Get maximum number of packages */

	max = strchr (data, '\n') + 1;
	stripNewline (max);
	maxPackages = atoi (max);

	if (maxPackages > 1)
	{
		gchar *msg;

		gtk_widget_show (W("globalProgress"));
		gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (W("globalProgress")), 0.0);
		msg = g_strdup_printf ("Preparing %d packages...", maxPackages);
		gtk_progress_bar_set_text (GTK_PROGRESS_BAR (W("globalProgress")), msg);
		g_free (msg);
	}

	installStarted = TRUE;
	return TRUE;
}


static gboolean
process_status (gchar *data)
{
	gchar *status;

	status = strchr (data, '\n') + 1;
	stripNewline (status);
	if (strlen (status) == 0)
		return TRUE;

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), status);
	return TRUE;
}


static gboolean
process_test (gchar *data)
{
	gchar *name, *str;

	name = strchr (data, '\n') + 1;
	stripNewline (name);
	if (strlen (name) == 0)
		return TRUE;

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	str = g_strdup_printf (_("Checking for %s..."), name);
	gtk_log_view_check_begin (GTK_LOG_VIEW (W("logview")), str);
	g_free (str);
	return TRUE;
}

static gboolean
process_searching(gchar *data)
{
	gtk_log_view_check_searching (GTK_LOG_VIEW(W("logview")));
	return TRUE;
}


static gboolean
process_progressbar (gchar *data)
{
	GList *lines = create_strlist (data);
	gdouble max, current;
	gchar *label;

	max = (gdouble) atoi ((gchar *) g_list_nth_data (lines, 1));
	current = (gdouble) atoi ((gchar *) g_list_nth_data (lines, 2));
	label = (gchar *) g_list_nth_data (lines, 3);

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);

	if (max == current)
	{
		/* The progress bar has reached it's end */
		if (progressStarted) gtk_log_view_set_progress (GTK_LOG_VIEW (W("logview")), 1.0);
		progressStarted = FALSE;

	} else
	{
		if (!progressStarted)
			gtk_log_view_add_progress (GTK_LOG_VIEW (W("logview")), label);
		gtk_log_view_set_progress (GTK_LOG_VIEW (W("logview")), current / max);
		progressStarted = TRUE;
	}

	free_strlist (lines);
	return TRUE;
}


static gboolean
process_fail (gchar *data)
{
	gchar *msg;

	msg = strchr (data, '\n') + 1;
	failed = TRUE;

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	stripNewline (msg);
	gtk_log_view_add_fail (GTK_LOG_VIEW (W("logview")), msg);

	return TRUE;
}


static gboolean
process_passtest (gchar *data)
{
	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	gtk_log_view_check_pass (GTK_LOG_VIEW (W("logview")));
	return TRUE;
}


static gboolean
process_failtest (gchar *data)
{
	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	gtk_log_view_check_fail (GTK_LOG_VIEW (W("logview")));
	return TRUE;
}


static gboolean
process_recommendtest (gchar *data)
{
	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	gtk_log_view_check_recommended (GTK_LOG_VIEW (W("logview")));
	return TRUE;
}


static gboolean
process_interact (gchar *data)
{
	GList *lines;
	gchar *name, *message, *defaultValue;

	lines = create_strlist (data);
	name = (gchar *) g_list_nth_data (lines, 1);
	message = (gchar *) g_list_nth_data (lines, 2);
	defaultValue = (gchar *) g_list_nth_data (lines, 3);

	if (startsWith (data, "INTERACT STRING\n"))
		interactionList = interaction_add (interactionList,
			name, message, defaultValue, INTERACT_STRING);

	else if (startsWith (data, "INTERACT PATH\n"))
		interactionList = interaction_add (interactionList,
			name, message, defaultValue, INTERACT_PATH);

	free_strlist (lines);
	return TRUE;
}


static gboolean
process_wait (gchar *data)
{
	interaction_execute (interactionList);
	return TRUE;
}


static gboolean
process_get (gchar *data)
{
	gchar *name;

	name = strchr (data, '\n') + 1;
	stripNewline (name);
	protocol_send ((gchar *) interaction_get (interactionList, name));
	return TRUE;
}


static gboolean
process_download_progress(gchar *data)
{
	GList *lines = create_strlist(data);
	char *rate = g_list_nth_data(lines, 1);
	char *progress = g_list_nth_data(lines, 2);
	/* char *time = g_list_nth_data(lines, 3); */

	char *markup = g_strdup_printf("%s%s%s%s", "<span foreground=\"grey\">", rate, _("kb/s"), "</span>");
	gtk_label_set_markup(GTK_LABEL(W("speed_label")), markup);
	g_free(markup);

	GtkProgressBar *progressbar = GTK_PROGRESS_BAR(W("download_progress"));

	int iprogress = atoi(progress);
	gtk_progress_bar_set_fraction(progressbar, (double) iprogress / 100);
	
	char *progress_text = g_strdup_printf(_("Download %d%% complete"), iprogress);
	gtk_progress_bar_set_text(progressbar, progress_text);
	g_free(progress_text);
	
	return TRUE;
}

static gboolean
process_download_start(gchar *data)
{
	GList *lines = create_strlist (data);
	char *url = g_list_nth_data (lines, 1);
	char *description = g_list_nth_data (lines, 2);

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 2);

	g_assert (url);
	g_assert (description);

	gul_ellipsizing_label_set_text (GUL_ELLIPSIZING_LABEL (W("url_label")), url);
	gtk_label_set_text (GTK_LABEL (W("speed_label")), "");

	free_strlist (lines);

	return TRUE;
}

static gboolean
process_download_prepare(gchar *data)
{
	GList *lines = create_strlist (data);
	char *state = g_list_nth_data(lines, 1);
	char *markup, *label;

	if (my_strcmp(state, "CONNECTING"))
		label = _("Connecting");
	else if (my_strcmp(state, "NEGOTIATING"))
		label = _("Negotiating");

	markup = g_strdup_printf("%s%s%s", "<span foreground='grey'>", label, "</span>");
	
	gtk_label_set_markup(GTK_LABEL(W("speed_label")), markup);
	
	g_free(markup);
	free_strlist(lines);
	return true;
}

static gboolean
process_download_finish(gchar *data)
{
	/* FIXME no-op for now */
	return TRUE;
}

static gboolean
process_display_summary (gchar *data)
{
	GList *lines = create_strlist (data);
	char *optype;
	char *result;
	char *redisplay;

	int i;

	optype = (char *) g_list_nth_data (lines, 1);
	result = (char *) g_list_nth_data (lines, 2);
	redisplay = (char *) g_list_nth_data (lines, 3);

	TRACE("optype=%s, result=%s, redisplay=%s\n", optype, result, redisplay);


	/* TODO: make this not suck */
	/* in particular, these strings are not good for translators */
	if (my_strcmp (optype, "install"))
	{
		int ok_packagecount;
		int menu_count;

		present_completed (COMPLETED_INSTALL);

		ok_packagecount = atoi ((char *) g_list_nth_data (lines, 4));
		menu_count = atoi ((char *) g_list_nth_data (lines, 4 + ok_packagecount + 1));

		if (ok_packagecount > 0)
		{
			gchar *middle, *msg;

			middle = (ok_packagecount > 1) ? g_strdup_printf (_("%d packages were"), ok_packagecount) : g_strdup (_("package was"));
			msg = g_strdup_printf (_("The following %s installed succesfully:"), middle);
			g_free (middle);

			gtk_log_view_clear (GTK_LOG_VIEW (W("logview")));
			gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), msg);
			g_free (msg);

			for (i = 0; i < ok_packagecount; i++)
		        {
				gchar *expanded_form;

				expanded_form = g_strdup_printf (" - %s", (gchar *) g_list_nth_data (lines, 5 + i));
				gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), expanded_form);
				g_free (expanded_form);
			}
		}

		if (menu_count > 0)
 		{
 			gchar *middle, *msg;

			middle = (menu_count > 1) ? g_strdup_printf (_("%d menu entries are"), menu_count) : g_strdup (_("menu entry is"));
			msg = g_strdup_printf (_("The following %s now available: "), middle);
			g_free (middle);

			if (ok_packagecount > 0) gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), "");

			gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), msg);
			g_free (msg);

			for (i = 4 + ok_packagecount + 2; i <= 4 + ok_packagecount + 1 + menu_count; i++)
			{
				char *menu_name;

				menu_name = desktop_file_get_name ((gchar *) g_list_nth_data (lines, i));
				if (menu_name)
				{
					gchar *expanded_form;
					
					expanded_form = g_strdup_printf (" - %s", menu_name);
					gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), expanded_form);
					g_free (expanded_form);
					g_free (menu_name);
				}
			}
		}

		if (my_strcmp (result, "failure")) {
			/* Failure adds an extra bit to the end of data, namely, the packages that failed.
			   therefore we can display packages that succeeded too. */
			char *failed_count_s = (char*) g_list_nth_data (lines, 4 + ok_packagecount + 1 + menu_count);
			int failed_count = atoi (failed_count_s);
			char *middle = (ok_packagecount > 1) ? g_strdup_printf ("%d packages", ok_packagecount) : g_strdup ("package");
			char *msg = g_strdup_printf (_("The following %s failed to install:"), middle);
			
			g_free (middle);

			gtk_log_view_add_status (GTK_LOG_VIEW (W("logview")), msg);
			g_free (msg);

			for (i = 5 + ok_packagecount + 1 + menu_count + 1; i <= 5 + ok_packagecount + 1 + menu_count + 1 + failed_count; i++)
			{
				gchar *expanded_form = g_strdup_printf (" - %s", (gchar *) g_list_nth_data (lines, i));
				gtk_log_view_add_status (GTK_LOG_VIEW(W("logview")), expanded_form);
				g_free (expanded_form);
			}
		}
	}
	else if (my_strcmp (optype, "uninstall"))
		printf ("FIXME\n");
	else if (my_strcmp (optype, "verify"))
	{
		if (my_strcmp (result, "success"))
			present_completed (COMPLETED_VERIFY);
		else
			present_completed (COMPLETED_PROBLEM);
	}
	
	summaryShown = TRUE;
	
	free_strlist (lines);
	return TRUE;
}


gboolean
process (gchar *data)
{
	gboolean result = TRUE;

	if (startsWith (data, "HELLO "))
		return process_hello (data);

	if (startsWith (data, "TERMINATE\n"))
		result = process_terminate ();

	else if (startsWith (data, "SWITCH GROUP\n"))
		result = process_switch_group (data);

	else if (startsWith (data, "IDENTIFY\n"))
		result = process_identify (data);

	else if (startsWith (data, "SERIES\n"))
		result = process_series (data);

	else if (startsWith (data, "STATUS\n"))
		result = process_status (data);

	else if (startsWith (data, "TEST\n"))
		result = process_test (data);

	else if (startsWith (data, "PROGRESSBAR\n"))
		result = process_progressbar (data);

	else if (startsWith (data, "FAIL\n"))
		result = process_fail (data);

	else if (startsWith (data, "FAILTEST\n"))
		result = process_failtest (data);

	else if (startsWith (data, "PASSTEST\n"))
		result = process_passtest (data);

	else if (startsWith (data, "RECOMMEND\n"))
		result = process_recommendtest (data);

	else if (startsWith (data, "INTERACT "))
		result = process_interact (data);

	else if (startsWith (data, "WAIT\n"))
		result = process_wait (data);

	else if (startsWith (data, "DISPLAY-SUMMARY\n"))
		result = process_display_summary (data);

	else if (startsWith (data, "DOWNLOAD-START\n"))
		result = process_download_start (data);
	
	else if (startsWith (data, "DOWNLOAD-PREPARE\n"))
		result = process_download_prepare (data);

	else if (startsWith (data, "DOWNLOAD-PROGRESS\n"))
		result = process_download_progress (data);

	else if (startsWith (data, "DOWNLOAD-FINISH\n"))
		result = process_download_finish (data);

	else if (startsWith (data, "SEARCHING\n"))
		result = process_searching (data);
	
	else if (startsWith (data, "GET\n"))
	{
		result = process_get (data);
		#ifdef DEBUG
		g_print ("\n");
		#endif
		return result;
	}

	else
	{
		show_error (W("install"),
			_("The backend sent an unknown command:\n%s"),
			data);
		exit (1);
	}

	protocol_send ("OK");
	/* #ifdef DEBUG
	g_print ("\n");
	#endif */
	return result;
}


/* Present an "installation completed" screen */
static void
present_completed (int what)
{
	gchar *msg, *tmp;

	presented = TRUE;

	gtk_notebook_set_current_page (GTK_NOTEBOOK (W("notebook1")), 1);
	pulse_stop ("globalProgress");
	gtk_widget_hide (W("globalProgress"));

	if (what == COMPLETED_PROBLEM)
		gtk_image_set_from_file (GTK_IMAGE (W("installImage")), lookup_file ("help.png"));
	else if (what == COMPLETED_INSTALL || what == COMPLETED_VERIFY || what == COMPLETED_UNKNOWN)
		gtk_image_set_from_file (GTK_IMAGE (W("installImage")), lookup_file ("package-ok.png"));
	gtk_widget_hide (W("prepareImage"));
	gtk_widget_show (W("installImage"));

	gtk_label_set_text (GTK_LABEL (W("name")), "");

	if (what == COMPLETED_PROBLEM)
		tmp = _("The operation failed to complete");
	else if (what == COMPLETED_INSTALL)
		tmp = _("Installation completed");
	else if (what == COMPLETED_VERIFY)
		tmp = _("Verification completed");
	else if (what == COMPLETED_UNKNOWN)
		tmp = _("Finished");
	msg = g_strdup_printf("<span weight=\"bold\" size=\"x-large\">%s</span>", tmp);
	gtk_label_set_markup (GTK_LABEL (W("title")), msg);
	g_free (msg);

	gtk_window_set_title (GTK_WINDOW (W("install")), tmp);

	gtk_widget_hide (W("hide"));
	gtk_widget_show (W("close"));
	gtk_button_set_label (GTK_BUTTON (W("close")), GTK_STOCK_OK);
	gtk_button_set_use_stock (GTK_BUTTON (W("close")), TRUE);
	gtk_widget_grab_default (W("close"));
	gtk_widget_grab_focus (W("close"));

	tmp = lookup_file ("package-open.png");
	if (!statusIcon) statusIcon = egg_status_icon_new();

	if (maxPackages == 1)
		msg = g_strdup_printf (_("Installation of the requested package is complete"));
	else
		msg = g_strdup_printf (_("Installation of %d packages is complete"), maxPackages);
	egg_status_icon_set_tooltip(statusIcon, msg, NULL);
	egg_status_icon_set_is_blinking (statusIcon, TRUE);
	egg_status_icon_set_from_file (statusIcon, tmp);
	g_signal_connect (G_OBJECT(statusIcon), "activate", G_CALLBACK (gtk_main_quit), NULL);

	g_free (tmp);
	g_free (msg);	
}

